# Requires SAGE (>= version 2.8), t3m, SnapPeaPython, and Magma
#
# Code for computing the Thurston norm and fibered faces of the manifold M of
#
# Nathan M Dunfield and Dinakar Ramakrishnan:
# Increasing the number of fibered faces of arithmetic hyperbolic 3-manifolds.
#
# In particular, it implements the techniques described in Section 6.7
# and 6.11 of that paper. 
#
# This code is in the public domain.  

import SnapPea
from t3m import *
import tempfile, string, os

from sage.matrix.matrix_generic_sparse import Matrix_sparse_from_rows

# First, we used the technique of Section 6.7 to create surfaces
# associated to elements in H_2(M, Z) = H^1(M, Z) for closed
# manifolds.

# The next function returns a matrix of generators of H^1 in terms of
# their values on the edges of manifold

def sparse_homology_matrix(manifold):
    VerticesOfFace = { F0 : (V1, V2, V3), F1 : (V2, V3, V0), F2 : (V3, V0, V1), F3 :
                       (V0, V1, V2) }

    n = len(manifold.Edges)
    rels = []
    # make matrix of relations
    V = VectorSpace(Rationals(), n, sparse=true)
    for tet in manifold.Tetrahedra:
        for face in TwoSubsimplices:
            rel = V(0)
            a, b, c = VerticesOfFace[face]
            e_ab = tet.Class[a | b]
            e_bc = tet.Class[b | c]
            e_ca = tet.Class[c | a]
            rel[e_ab.index()] = rel[e_ab.index()]  + e_ab.orientation_with_respect_to(tet, a, b)
            rel[e_bc.index()] = rel[e_bc.index()] + e_bc.orientation_with_respect_to(tet, b, c)
            rel[e_ca.index()] = rel[e_ca.index()] + e_ca.orientation_with_respect_to(tet, c, a)

            rels.append(rel)

    return Matrix_sparse_from_rows(rels).change_ring(Integers())

def first_cohomology(manifold):
    M = sparse_homology_matrix(manifold).transpose().dense_matrix()
    return M.kernel()

def get_signed_weight(cohom, tet, a, b):
    sign = tet.Class[a | b].orientation_with_respect_to(tet, a, b)
    return sign * cohom[tet.Class[a | b].index()]

# Compute the euler characteristic of the normal surface associated to
# the piecewise affine map to the circle given by the particular
# cohomology class.

def euler_of_dual_surface(M, C):   #manifold, cohomology class
    V = 0
    for w in C:
        V = V + abs(w)

    E , F = 0, 0
    for tet in M.Tetrahedra:
        E_change = 0
        for edge in OneSubsimplices:
            ind = tet.Class[edge].index()
            E_change = E_change + abs(C[ind])

        E = E + 0.5 * E_change

        a = get_signed_weight(C, tet, V0, V1)
        b = get_signed_weight(C, tet, V0, V2)
        c = get_signed_weight(C, tet, V0, V3)

        trans = [0, a, b, c]
        F = F + (max(trans) - min(trans))

    ans = int(F - E + V)
    assert ans == F - E + V
    return ans

#--------------------------------------------------------------------
#
# Constructing the fundamental group directly from the triangulation.
#
#-------------------------------------------------------------------

def build_magma_group_string(generators, relations):
    expanded_rels = [ "*".join(rel) for rel in relations]
    return "Group< " + ", ".join(generators) + " | " + ", ".join(expanded_rels) + ">"

def magma_fp_group(generators, relations):
    G = magma.new(build_magma_group_string(generators, relations))
    G.AssignNames(generators)
    return G

def fundamental_group_presentation(manifold):
    assert len(manifold.Vertices) == 1
    
    VerticesOfFace = { F0 : (V1, V2, V3), F1 : (V2, V3, V0), F2 : (V3, V0, V1), F3 :
                       (V0, V1, V2) }

    n = len(manifold.Edges)
    gens = ["a%d" % i for i in range(n)]
    rels = []
    for tet in manifold.Tetrahedra:
        for face in TwoSubsimplices:
            a, b, c = VerticesOfFace[face]
            e_ab = tet.Class[a | b]
            e_bc = tet.Class[b | c]
            e_ca = tet.Class[c | a]
            rels.append([
                gens[e_ab.index()]  + "^" + repr(e_ab.orientation_with_respect_to(tet, a, b)), 
                gens[e_bc.index()]  + "^" +  repr(e_bc.orientation_with_respect_to(tet, b, c)), 
                gens[e_ca.index()] + "^" + repr(e_ca.orientation_with_respect_to(tet, c, a))]
                )

    return (gens, rels)

def fundamental_group(manifold):
    return magma_fp_group( *fundamental_group_presentation(manifold) )

#------------------------------------------------
#
#  Dealing with Alexander polynomials
#
#------------------------------------------------

magma.load("alexander.magma")

# Getting a polynomial out of Magma into Sage

magma.eval("""
DeconstructPolynomial := function(p)
   return [Exponents(m) : m in Monomials(p)], Coefficients(p);
end function;
""")

def sage_poly_from_magma(poly):
    exponents, coeffs = poly.DeconstructPolynomial(nvals=2)
    exponents, coeffs = exponents.sage(), coeffs.sage()
    n = len(exponents[0])
    R = PolynomialRing(Rationals(), list(string.lowercase[:n]))
    gens = R.gens()
    def make_term(exp, c):
        ans = c
        for i in range(len(exp)):
            ans = ans * gens[i]**exp[i]
        return ans

    return sum(
        [ make_term(exponents[i], coeffs[i]) for i in range(len(exponents))])

# Computing the Alexander polynomial of the given Magma finitely
# presented group.

def alexander_polynomial(group, cohom=None):
    if not cohom:
        p, R = group.AlexanderPolynomial(nvals = 2)
        n = R.Rank().sage()
    else:
        n = len(cohom)
        p, R = group.AlexanderPolynomialFromCohomologyClasses([ map(int, list(c)) for c in cohom], nvals = 2)

    return  AlexanderPolynomial(sage_poly_from_magma(p))


# A class for keeping track of Alexander polynomials; in can compute
# the Alexander norm described in Section 6.7, and give information
# about the Newton polytope and the dual Alexander norm ball using the
# Polymake additional package

class AlexanderPolynomial:
    def __init__(self, poly):
        self.n = len(poly.parent().gens())        
        self.Exponents = [list(x) for x in poly.exponents()]
        self.Coefficients = poly.coefficients()
        self.Polynomial = poly

        self.Matrix = MatrixSpace(Integers(), len(self.Exponents), self.n )(self.Exponents)
        if n > 1:
            self.add_newton()

    def __repr__(self):
        return "<AlexPoly : %s>" % self.Polynomial

    def complete_norm(self, vector):
        vals = list(self.Matrix * VectorSpace(Rationals(), self.n)(vector))
        m, M = min(vals), max(vals)
        return M - m,  self.Coefficients[vals.index(M)], vals.count(M) == 1

    def norm(self, vector):
        return self.complete_norm(vector)[0]

    def should_fiber(self, vector):
        norm, coeff, generic = self.complete_norm(vector)
        return  generic and abs(coeff) == 1


    def add_newton(self):
        P = self.Polynomial.newton_polytope()
        n = self.n
        V = VectorSpace(Rationals(), n)

        # Find the dual polytope (assumes newton polytope is symmetric)

        directions = map(V,  list(set([tuple(f[1:]) for f in P.facets()])))
        dual_verts = [ (1/self.norm(v)) * v for v in directions] +  [- (1/self.norm(v)) * v for v in directions]
        D = polymake.convex_hull([ [1] + list(v) for v in dual_verts])

        self.NewtonPolytope = P
        self.DualPolytope = D

    def dual_vertices(self):
       rawverts =  [ v[1:] for v in self.DualPolytope.vertices()]
       clear_denom_factor = LCM( [  LCM( [x.denominator() for x in v]) for v in rawverts] )
       return [clear_denom_factor * v for v in rawverts]

    def vertices(self):
       rawverts =  [ v[1:] for v in self.NewtonPolytope.vertices()]
       centroid = (QQ(len(rawverts))**-1) * sum(rawverts)
       rawverts = [v - centroid for v in rawverts]
       clear_denom_factor = LCM( [  LCM( [x.denominator() for x in v]) for v in rawverts] )
       return [clear_denom_factor * v for v in rawverts]

# Next we add methods to the t3m triangulation class for computing
# cohomology and Alexander polynomials.
                                          
def compute_cohomology(self):
    K = first_cohomology(self)
    basis = K.basis()
    self.CohomologyBasisMatrix = MatrixSpace(Integers(), K.dimension(), K.degree())(basis)
    self.BettiNumber = K.dimension()

def compute_alexander_polynomial(self):
    self.compute_cohomology()
    self.AlexanderPolynomial = alexander_polynomial(fundamental_group(self), list(self.CohomologyBasisMatrix))

def alexander_norm(self, vector):
    return self.AlexanderPolynomial.norm(vector)

def pseudo_thurston_norm(self, vector):
    return - euler_of_dual_surface(self, self.cohomology_class_to_weights(vector))

def pseudo_norms_of_basis(self):
    return [self.pseudo_thurston_norm(v) for v in MatrixSpace(Integers(), self.BettiNumber)(1)]

def cohomology_class_to_weights(self, vector):
    return list(VectorSpace(Rationals(), self.BettiNumber)(vector) * self.CohomologyBasisMatrix)

Mcomplex.compute_cohomology = compute_cohomology
Mcomplex.compute_alexander_polynomial = compute_alexander_polynomial
Mcomplex.alexander_norm = alexander_norm
Mcomplex.pseudo_thurston_norm = pseudo_thurston_norm
Mcomplex.pseudo_norms_of_basis = pseudo_norms_of_basis
Mcomplex.cohomology_class_to_weights = cohomology_class_to_weights

#---------------------------
#
# Some helper functions for creating useful triangulations from SnapPea
# manifolds.
#
#--------------------------

def SnapPea_to_Mcomplex(M):
    file = tempfile.mktemp()
    M.save(file)
    N = read_SnapPea_file(file)
    os.remove(file)
    return N

def find_small_finite_triangulation(M, trys = 100):
    N = M.make_finite_triangulation()
    best_manifold = N.clone()
    for i in range(trys):
        N.randomize()
        if N.get_num_tetrahedra() < best_manifold.get_num_tetrahedra():
            best_manifold = N.clone()

    return best_manifold

def example_manifold():
    M = SnapPea.get_manifold("9n6(0,1)")
    return SnapPea_to_Mcomplex(find_small_finite_triangulation(M))

#----------------------------------
#
#  Next we implement the fibering test described in Section 6.11
#
#----------------------------------

def edge_orientations(tet):
     return [tet.Class[edge].orientation_with_respect_to(tet, Tail[edge], Head[edge]) for
             edge in OneSubsimplices]

def edge_weights(tet, cohom_class):
    signs = edge_orientations(tet)
    return [signs[i] * cohom_class[tet.Class[OneSubsimplices[i]].index()] for i in range(6)]

# Checks if the surface falls into the simple case depicted in Figure 6.13

def is_good_candidate(self, cohom_class):
    C = self.cohomology_class_to_weights(cohom_class)
    if max(map(abs, C)) > 2:
        return False

    for tet in self.Tetrahedra:
        abs_edge_wts = map(abs, edge_weights(tet, C))
        if abs_edge_wts.count(2) > 1:
            return False

    return True

# Works only assuming that the surface satisfies is_good_candidate above

def comp_spline_fundamental_group_for_magma(self, cohomology_class):
    assert(is_good_candidate(self, cohomology_class))
    cohom = self.cohomology_class_to_weights(cohomology_class)
    ans =  "Group< "+ repr(["e%d" % i for i in range(len(cohom)) if cohom[i] == 0])[1:-1].replace("'", "")+" | "
    VertsInFaces = [ ZeroSubsimplices[:i]+ ZeroSubsimplices[i + 1:] for i in range(4)]

    relations = []
    for tet in self.Tetrahedra:
        for verts in VertsInFaces:
            # We don't want to double-print everything
            v1,v2,v3 = verts
            face = v1|v2|v3
            if tet.Neighbor.get(face).Index > tet.Index:
                edges =[ v1 | v2, v2 | v3, v3 | v1]
                signs = [tet.Class[edge].orientation_with_respect_to(tet, verts[k], verts[(k + 1)%3]) for k, edge in enumerate(edges)]
                sign_symbol = {-1:"^-1", 1:"^1"}
                signs = [sign_symbol[s] for s in signs]
                indices = [tet.Class[edge].index() for edge in edges]
                wts = [cohom[i] for i in indices]
                if wts == [0,0,0]:
                    relations.append( "e%d%s * e%d%s * e%d%s" % (
                        indices[0],signs[0], indices[1],  signs[1], indices[2],signs[2]) )

    # now we give the relations for each edge of weight 2:

    two_edges = [self.Edges[i] for i in range(len(cohom)) if abs(cohom[i]) == 2]
    for edge in two_edges:
        relation = []
        a = edge.get_arrow()
        for i in range(edge.valence()):
            tet = a.Tetrahedron
            C = tet.Class[comp(a.Edge)]
            relation.append(  ("%s" % C)[:-6] + sign_symbol[C.orientation_with_respect_to(tet, a.tail(), a.head())] )
            a.next()
        relations.append("*".join(relation))

    return ans +  ", ".join(relations) + ">"

def comp_spline_fundamental_group_string_for_gap(self, cohomology_class):
    assert(is_good_candidate(self, cohomology_class))
    cohom = self.cohomology_class_to_weights(cohomology_class)
    genstrings = ["e%d" % i for i in range(len(cohom)) if cohom[i] == 0];
    
    ans =  "CallFuncList(function() local " + ", ".join(["F"] + genstrings)  + ";"
    ans += " F := FreeGroup(%d); " % len(genstrings)
    for i in range(len(genstrings)):
        ans += genstrings[i] + " := F.%d;  "  % (i + 1)
        
    VertsInFaces = [ ZeroSubsimplices[:i]+ ZeroSubsimplices[i + 1:] for i in range(4)]

    relations = []
    for tet in self.Tetrahedra:
        for verts in VertsInFaces:
            # We don't want to double-print everything
            v1,v2,v3 = verts
            face = v1|v2|v3
            if tet.Neighbor.get(face).Index > tet.Index:
                edges =[ v1 | v2, v2 | v3, v3 | v1]
                signs = [tet.Class[edge].orientation_with_respect_to(tet, verts[k], verts[(k + 1)%3]) for k, edge in enumerate(edges)]
                sign_symbol = {-1:"^-1", 1:"^1"}
                signs = [sign_symbol[s] for s in signs]
                indices = [tet.Class[edge].index() for edge in edges]
                wts = [cohom[i] for i in indices]
                if wts == [0,0,0]:
                    relations.append( "e%d%s * e%d%s * e%d%s" % (
                        indices[0],signs[0], indices[1],  signs[1], indices[2],signs[2]) )

    # now we give the relations for each edge of weight 2:

    two_edges = [self.Edges[i] for i in range(len(cohom)) if abs(cohom[i]) == 2]
    for edge in two_edges:
        relation = []
        a = edge.get_arrow()
        for i in range(edge.valence()):
            tet = a.Tetrahedron
            C = tet.Class[comp(a.Edge)]
            relation.append(  ("%s" % C)[:-6] + sign_symbol[C.orientation_with_respect_to(tet, a.tail(), a.head())] )
            a.next()
        relations.append("*".join(relation))

    return    ans + "return F/[" + ", ".join(relations) + "]; end, [])"

#-----------------------------------------------------------------------------
#
#  Verifying the needed properties of M and N asserted in Section 6.7 and 6.11
#
#---------------------------------------------------------------------------

# With respect to the current triangulation, check which vertices of the
# Alexander norm ball also agree with the Thurston norm.

def check_vertices(manifold):
    return [
        v for v in manifold.AlexanderPolynomial.dual_vertices()
        if manifold.pseudo_thurston_norm(v) == manifold.alexander_norm(v)]

# Check the claims about M from Section 6.7

def main_check_M():
    M = read_SnapPea_file("Mgood")
    M.compute_alexander_polynomial()
    P = M.AlexanderPolynomial
    print P
    print factor(P.Polynomial)
    print P.vertices()
    print P.dual_vertices()
    print [P.complete_norm(v) for v in P.vertices()]
    print [P.complete_norm(v) for v in P.dual_vertices()]
    print check_vertices(M)

#  Ngood created by
#
# sage: Nraw = SnapPea.get_manifold("Nraw")
# sage: NN = SnapPea_to_Mcomplex(find_small_finite_triangulation(Nraw))
# sage: write_SnapPea_file(NN, "Ngood")

# Code used for checking the claims of Section 6.11

def main_check_N():
    N = read_SnapPea_file("Ngood")
    N.compute_cohomology()
    G = magma(comp_spline_fundamental_group_for_magma(N, [1]))
    H = G.Simplify()
    H.AssignNames("abcd")
    print H
    G = gap(comp_spline_fundamental_group_string_for_gap(N, [1]))
    H = G.SimplifiedFpGroup()
    print H.GeneratorsOfGroup(), H.RelatorsOfFpGroup()
    
    

# Searching code used for M

def search_for_good_triangulation(snappea_mfld, trys = 100):
    N = snappea_mfld.make_finite_triangulation()
    for i in range(trys):
        for j in range( random.choice([1,2,4,8,16,32]) ):
            N.randomize()
        NN = SnapPea_to_Mcomplex(N)
        if len(NN.Vertices) == 1:
            NN.compute_alexander_polynomial()
            ans = check_vertices(NN)
            print i, j, ans
            if len(ans) > 0:
                N.save("good_%d" % i)

# Searching code used for N

def search_for_good_triangulation2(snappea_mfld, trys = 100):
    N = snappea_mfld.make_finite_triangulation()
    for i in range(trys):
        for j in range( random.choice([1,2,4,8,16,32]) ):
            N.randomize()
        NN = SnapPea_to_Mcomplex(N)
        if len(NN.Vertices) == 1:
            NN.compute_cohomology()
            if NN.pseudo_norms_of_basis() == [2] and is_good_candidate(NN, [1]):
                N.save("good_%d" % i)
                print i, comp_spline_fundamental_group_for_magma(NN, [1])


"""

The output of the various checks in this file

sage: attach thurston_norm.py
sage: main_check_M()
<AlexPoly : 65536*a^4*b^4*c^4 - 16384*a^4*b^4*c^3 - 16384*a^4*b^3*c^4 + 1536*a^4*b^4*c^2 + 3072*a^4*b^3*c^3 - 16384*a^3*b^4*c^3 + 1536*a^4*b^2*c^4 - 16384*a^3*b^3*c^4 - 64*a^4*b^4*c - 192*a^4*b^3*c^2 + 3072*a^3*b^4*c^2 - 192*a^4*b^2*c^3 + 268288*a^3*b^3*c^3 - 64*a^4*b*c^4 + 3072*a^3*b^2*c^4 + a^4*b^4 + 4*a^4*b^3*c - 192*a^3*b^4*c + 6*a^4*b^2*c^2 - 49728*a^3*b^3*c^2 + 1536*a^2*b^4*c^2 + 4*a^4*b*c^3 - 49728*a^3*b^2*c^3 + 3072*a^2*b^3*c^3 + a^4*c^4 - 192*a^3*b*c^4 + 1536*a^2*b^2*c^4 + 4*a^3*b^4 + 3088*a^3*b^3*c - 192*a^2*b^4*c + 6168*a^3*b^2*c^2 - 49728*a^2*b^3*c^2 + 3088*a^3*b*c^3 - 49728*a^2*b^2*c^3 + 4*a^3*c^4 - 192*a^2*b*c^4 - 64*a^3*b^3 + 6*a^2*b^4 - 192*a^3*b^2*c + 6168*a^2*b^3*c - 64*a*b^4*c - 192*a^3*b*c^2 + 405540*a^2*b^2*c^2 - 192*a*b^3*c^2 - 64*a^3*c^3 + 6168*a^2*b*c^3 - 192*a*b^2*c^3 + 6*a^2*c^4 - 64*a*b*c^4 - 192*a^2*b^3 + 4*a*b^4 - 49728*a^2*b^2*c + 3088*a*b^3*c - 49728*a^2*b*c^2 + 6168*a*b^2*c^2 - 192*a^2*c^3 + 3088*a*b*c^3 + 4*a*c^4 + 1536*a^2*b^2 - 192*a*b^3 + b^4 + 3072*a^2*b*c - 49728*a*b^2*c + 4*b^3*c + 1536*a^2*c^2 - 49728*a*b*c^2 + 6*b^2*c^2 - 192*a*c^3 + 4*b*c^3 + c^4 + 3072*a*b^2 - 64*b^3 + 268288*a*b*c - 192*b^2*c + 3072*a*c^2 - 192*b*c^2 - 64*c^3 - 16384*a*b + 1536*b^2 - 16384*a*c + 3072*b*c + 1536*c^2 - 16384*b - 16384*c + 65536>
(16*a*b*c - a*b - a*c - b - c + 16)^4
[(-2, -2, -2), (-2, -2, 2), (-2, 2, -2), (2, -2, 2), (2, 2, -2), (2, 2, 2)]
[(-1, 0, 0), (-1, 1, 1), (0, -1, 0), (0, 0, -1), (0, 0, 1), (0, 1, 0), (1, -1, -1), (1, 0, 0)]
[(24, 65536, True), (24, 1, True), (24, 1, True), (24, 1, True), (24, 1, True), (24, 65536, True)]
[(4, 1, False), (4, 65536, False), (4, 1, False), (4, 1, False), (4, 65536, False), (4, 65536, False), (4, 1, False), (4, 65536, False)]
[(-1, 0, 0), (1, 0, 0)]

sage: main_check_N()
Finitely presented group on 4 generators
Relations
c * d * b * a^-1 * b^-1 * a * c^-1 * d^-1 = Id($)
[ f3, f4, f5, f7 ] [ f5*f7*f4*f3^-1*f4^-1*f3*f5^-1*f7^-1 ]

"""
